/****************************************************************************
 *                                                                          *
 * File    : Export.c                                                       *
 *                                                                          *
 * Purpose : Add-In for Pelles C - export C source as HTML, PDF file.       *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-16  Created                                              *
 *           05-03-18  Added support for PDF files.                         *
 *           05-03-21  Added compression to PDF files (thanks Timppa!)      *
 *           05-12-23  Robert Wishlaw enclosed source with <pre> ... </pre>,*
 *                     allowing <br> to be discarded, and replaced "class"  *
 *                     with "id", and replaced "span" with "code" allowing  *
 *                     "&nbsp;" to be replaced with a space.                *
 *           06-01-02  Robert Wishlaw reverted "id" to class.               *
 *                                                                          *
 ****************************************************************************/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <commdlg.h>
#include <addin.h>
#include <tchar.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef USE_PDF_COMPRESSION
// We use ZLIB for PDF compresion.
#include "zlib-1.2.2\zlib.h"
#endif /* USE_PDF_COMPRESSION */

#define NELEMS(a)  (sizeof(a) / sizeof(a[0]))

#define COOKIE_COMMENT       0x0001
#define COOKIE_PREPROCESSOR  0x0002
#define COOKIE_EXT_COMMENT   0x0004
#define COOKIE_STRING        0x0008
#define COOKIE_CHAR          0x0010

enum {
    COLOR_TEXT = 0,
    COLOR_KEYWORD = 1,
    COLOR_COMMENT = 2,
    COLOR_NUMBER = 3,
    COLOR_STRING = 4,
    COLOR_PREPROCESSOR = 5,
    COLOR_OPERATOR = 6,
};

typedef struct _PARSEPOINT *PPARSEPOINT;
typedef struct _PARSEPOINT {
    int iChar;                      // Column position where text changes...
    int iColor;                     // ...to this color index.
} PARSEPOINT;

typedef enum {
    PDF_TYPE_CATALOG = 1,           // Document Catalog.
    PDF_TYPE_OUTLINES,              // Outline Dictionary.
    PDF_TYPE_PAGES,                 // Pages Dictionary.
    PDF_TYPE_PAGE,                  // Page Node.
    PDF_TYPE_CONTENT,               // Page Content Node.
    PDF_TYPE_PROCSET,               // Procedure Set Node (obsolete from PDF 1.4 -- only for compatibility).
    PDF_TYPE_FONT,                  // Font Node.
    PDF_TYPE_INFO,                  // Info Node.
} PDF_TYPE;

typedef struct tagPDFOBJ {
    struct tagPDFOBJ *pNext;        // Pointer to next node.
    PDF_TYPE nType;                 // Object type.
    int nNumber;                    // Object number.
    long nFileOffset;               // Object file offset (for cross-reference).
    PSTR pszData;                   // Pointer to data or NULL.
    //
    int iCol;                       // Column position for PDF_TYPE_CONTENT nodes.
    int iLastColor;                 // Last emitted color index or -1. File size optimization.
} PDFOBJ, *PPDFOBJ;

typedef struct tagPDFFILE {
    int cObjects;                   // Number of objects.
    PPDFOBJ pObjHead;               // Linked list of objects.
    PPDFOBJ pObjProc;               // Procedure Set node.
    PPDFOBJ pObjFont;               // Font node.
} PDFFILE, *PPDFFILE;

// Private command ID.
#define ID_EXPORT_HTML  1
#define ID_EXPORT_PDF  2

// PDF limits.
#define PDF_MAXLINES  58
#define PDF_MAXCHARS  80

// Locals.
static HANDLE g_hmod = NULL;
static HWND g_hwndMain = NULL;

// C keywords + M$ junk.
static PCTSTR apszKeywords[] = {
    _T("_Bool"),
    _T("_Complex"),
    _T("_Imaginary"),
    _T("_Pragma"),
    _T("__asm"),
    _T("__cdecl"),
    _T("__declspec"),
    _T("__except"),
    _T("__fastcall"),
    _T("__finally"),
    _T("__func__"),
    _T("__inline"),
    _T("__int16"),
    _T("__int32"),
    _T("__int64"),
    _T("__int8"),
    _T("__leave"),
    _T("__stdcall"),
    _T("__try"),
    _T("_asm"),
    _T("_cdecl"),
    _T("_fastcall"),
    _T("_inline"),
    _T("_stdcall"),
    _T("auto"),
    _T("break"),
    _T("case"),
    _T("char"),
    _T("const"),
    _T("continue"),
    _T("default"),
    _T("do"),
    _T("double"),
    _T("else"),
    _T("enum"),
    _T("extern"),
    _T("float"),
    _T("for"),
    _T("goto"),
    _T("if"),
    _T("inline"),
    _T("int"),
    _T("long"),
    _T("register"),
    _T("restrict"),
    _T("return"),
    _T("short"),
    _T("signed"),
    _T("sizeof"),
    _T("static"),
    _T("struct"),
    _T("switch"),
    _T("typedef"),
    _T("union"),
    _T("unsigned"),
    _T("void"),
    _T("volatile"),
    _T("while")
};

#define DEFINE_BLOCK(pos, colorindex)  \
    if (aPoints != NULL) \
    {\
        if (*pcPoints == 0 || aPoints[(*pcPoints)-1].iChar <= (pos)) \
        { \
            aPoints[(*pcPoints)].iChar = (pos);\
            aPoints[(*pcPoints)].iColor = (colorindex);\
            (*pcPoints)++; \
        } \
    }

// Static function prototypes.
static void ExportHTML(PCTSTR, PCTSTR);
static BOOL HTML_WriteText(FILE *, PCSTR, int, PINT);
static void ExportPDF(PCSTR, PCTSTR);
static PPDFFILE PDF_Init(void);
static void PDF_Done(PPDFFILE);
static PPDFOBJ PDF_NewObject(PPDFFILE, PDF_TYPE);
static PPDFOBJ PDF_FindObject(PPDFOBJ, PDF_TYPE);
static int PDF_AppendData(PPDFOBJ, PCSTR);
static PPDFOBJ PDF_StartPage(PPDFFILE);
static int PDF_EndPage(PPDFOBJ);
static int PDF_NewLine(PPDFOBJ);
static int PDF_WriteFile(PPDFFILE, FILE *);
static int PDF_EmitText(PPDFOBJ, PCSTR, int);
static int PDF_EmitColor(PPDFOBJ, int);
static UINT C_Parser(UINT, PCTSTR, int, PARSEPOINT [], PINT);
static int IsKeyword(PCTSTR [], int, PCTSTR, int);
static BOOL IsNumber(PCTSTR, int);

/****************************************************************************
 *                                                                          *
 * Function: DllMain                                                        *
 *                                                                          *
 * Purpose : DLL entry and exit procedure.                                  *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-16  Created                                              *
 *                                                                          *
 ****************************************************************************/

BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved)
{
    switch (dwReason)
    {
        case DLL_PROCESS_ATTACH:
            // Save module handle, in case we need it.
            g_hmod = hDLL;
            return TRUE;

        case DLL_PROCESS_DETACH:
            return TRUE;

        default:
            return TRUE;
    }
}

/****************************************************************************
 *                                                                          *
 * Function: AddInMain                                                      *
 *                                                                          *
 * Purpose : Add-in entry and exit procedure.                               *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-16  Created                                              *
 *                                                                          *
 ****************************************************************************/

ADDINAPI BOOL WINAPI AddInMain(HWND hwnd, ADDIN_EVENT eEvent)
{
    switch (eEvent)
    {
        case AIE_APP_CREATE:
        {
            ADDIN_ADD_COMMAND AddCmd = {0};

            // Remember main window handle.
            g_hwndMain = hwnd;

            // Add choice to source menu.
            AddCmd.cbSize = sizeof(AddCmd);
            AddCmd.pszText = _T("Export C source as HTML file...");
            AddCmd.hIcon = NULL;
            AddCmd.id = ID_EXPORT_HTML;
            AddCmd.idMenu = AIM_MENU_SOURCE;
            if (!AddIn_AddCommand(hwnd, &AddCmd))
                return FALSE;

            // Add choice to source menu.
            AddCmd.cbSize = sizeof(AddCmd);
            AddCmd.pszText = _T("Export C source as PDF file...");
            AddCmd.hIcon = NULL;
            AddCmd.id = ID_EXPORT_PDF;
            AddCmd.idMenu = AIM_MENU_SOURCE;
            if (!AddIn_AddCommand(hwnd, &AddCmd))
                return FALSE;

            return TRUE;
        }

        case AIE_APP_DESTROY:
            // Forget main window handle.
            g_hwndMain = NULL;
            // Remove choices from the source menu.
            return AddIn_RemoveCommand(hwnd, ID_EXPORT_HTML) &&
                   AddIn_RemoveCommand(hwnd, ID_EXPORT_PDF);

        default:
            return TRUE;
    }
}

/****************************************************************************
 *                                                                          *
 * Function: AddInCommandEx                                                 *
 *                                                                          *
 * Purpose : Add-in command handler.                                        *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-16  Created                                              *
 *                                                                          *
 ****************************************************************************/

ADDINAPI void WINAPI AddInCommandEx(int idCmd, LPCVOID pcvData)
{
    switch (idCmd)
    {
        case ID_EXPORT_HTML:
        {
            OPENFILENAME ofn = {0};
            TCHAR szFilename[260] = _T("");

            // Get name of HTML file to save.
            ofn.lStructSize = sizeof(ofn);
            ofn.hwndOwner = g_hwndMain;
            ofn.lpstrFilter = _T("HTML files (*.htm;*.html)\0*.htm;*.html\0");
            ofn.nFilterIndex = 1;
            ofn.lpstrFile = szFilename;
            ofn.nMaxFile = NELEMS(szFilename);
            ofn.lpstrDefExt = _T("htm");
            if (GetSaveFileName(&ofn))
            {
                // Get handle of the current document window - must be a source window.
                HWND hwndDoc = AddIn_GetActiveDocument(g_hwndMain);
                if (IsWindow(hwndDoc))
                {
                    // Get total length and allocate buffer.
                    int cchText = AddIn_GetSourceTextLength(hwndDoc);
                    PTSTR pchText = malloc(cchText + 1);
                    if (pchText != NULL)
                    {
                        // Get all text into buffer and output.
                        if (AddIn_GetSourceText(hwndDoc, pchText, cchText + 1) > 0)
                            ExportHTML(szFilename, pchText);

                        // Free buffer.
                        free(pchText);
                    }
                }
            }
            break;
        }

        case ID_EXPORT_PDF:
        {
            OPENFILENAME ofn = {0};
            TCHAR szFilename[260] = _T("");

            // Get name of PDF file to save.
            ofn.lStructSize = sizeof(ofn);
            ofn.hwndOwner = g_hwndMain;
            ofn.lpstrFilter = _T("PDF files (*.pdf)\0*.pdf\0");
            ofn.nFilterIndex = 1;
            ofn.lpstrFile = szFilename;
            ofn.nMaxFile = NELEMS(szFilename);
            ofn.lpstrDefExt = _T("pdf");
            if (GetSaveFileName(&ofn))
            {
                // Get handle of the current document window - must be a source window.
                HWND hwndDoc = AddIn_GetActiveDocument(g_hwndMain);
                if (IsWindow(hwndDoc))
                {
                    // Get total length and allocate buffer.
                    int cchText = AddIn_GetSourceTextLength(hwndDoc);
                    PTSTR pchText = malloc(cchText + 1);
                    if (pchText != NULL)
                    {
                        // Get all text into buffer and output.
                        if (AddIn_GetSourceText(hwndDoc, pchText, cchText + 1) > 0)
                            ExportPDF(szFilename, pchText);

                        // Free buffer.
                        free(pchText);
                    }
                }
            }
            break;
        }
    }
}

/****************************************************************************
 *                                                                          *
 * Function: ExportHTML                                                     *
 *                                                                          *
 * Purpose : Export C source code as formatted HTML file.                   *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-16  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void ExportHTML(PCTSTR pszFilename, PCTSTR pszText)
{
    FILE *fd = fopen(pszFilename, "w");
    if (fd != NULL)
    {
        PCTSTR pchLine, pchNext;
        UINT uCookie;
        int iCol = 0;

        // Write header.
        if (fputs(
            "<html>\n"
            "<head>\n"
            "<meta name=\"GENERATOR\" content=\"Export Add-In - www.smorgasbordet.com\">\n"
            "</head>\n"
            "<body>\n"
            "<style type=\"text/css\">\n"
            "code.C0 { color:#000000; background-color:#FFFFFF; }\n"  /* COLOR_TEXT */
            "code.C1 { color:#0000FF; background-color:#FFFFFF; font-weight:bold; }\n"  /* COLOR_KEYWORD */
            "code.C2 { color:#008000; background-color:#FFFFFF; }\n"  /* COLOR_COMMENT */
            "code.C3 { color:#FF00FF; background-color:#FFFFFF; }\n"  /* COLOR_NUMBER */
            "code.C4 { color:#800080; background-color:#FFFFFF; }\n"  /* COLOR_STRING */
            "code.C5 { color:#800000; background-color:#FFFFFF; }\n"  /* COLOR_PREPROCESSOR */
            "code.C6 { color:#0000FF; background-color:#FFFFFF; }\n"  /* COLOR_OPERATOR */
            "code {  color:#000000; background-color:#FFFFFF; font-family:monospace; font-size: 10pt; }\n"
            "body { background:#FFFFFF; }\n"
            "</style>\n"
			"<pre>\n", fd) < 0) goto error;

        // Walk through the source file.
        uCookie = 0;
        for (pchLine = pchNext = pszText; *pchLine != '\0'; pchLine = pchNext)
        {
            int cPoints = 0, iPoint, iColor;
            PCTSTR pchCode, pchText;
            static PARSEPOINT aPoints[4096];

            // Find the end of the current line.
            while (*pchNext != '\0' && *pchNext++ != '\n')
                ;

            // Parse the current line, to get syntax color points.
            uCookie = C_Parser(uCookie, pchLine, (pchNext - pchLine), aPoints, &cPoints);

            // Walk through the line again, looking for color changes.
            iPoint = 0; iColor = COLOR_TEXT;
            for (pchCode = pchText = pchLine; pchText < pchNext; pchText++)
            {
                int iChar = pchText - pchLine;

                // Hit EOL?
                if (*pchText == '\n')
                    break;

                // Hit a color change?
                if (iPoint < cPoints && aPoints[iPoint].iChar == iChar)
                {
                    // Anything to print before the color change?
                    if (pchCode < pchText)
                    {
                        if (fprintf(fd, "<code class=\"C%d\">", iColor) < 0) goto error;
                        if (!HTML_WriteText(fd, pchCode, (pchText - pchCode), &iCol)) goto error;
                        if (fputs("</code>", fd) < 0) goto error;
                        pchCode = pchText;
                    }

                    // In case we have multiple changes in the same spot...
                    while (iPoint < cPoints && aPoints[iPoint].iChar == iChar)
                        iColor = aPoints[iPoint++].iColor;
                }
            }

            // Handle any dangling text.
            if (pchCode < pchText)
            {
                if (fprintf(fd, "<code class=\"C%d\">", iColor) < 0) goto error;
                if (!HTML_WriteText(fd, pchCode, (pchText - pchCode), &iCol)) goto error;
                if (fputs("</code>", fd) < 0) goto error;
            }

            // Why not a line break in the HTML file too?
            // if (fputs("<br>\n", fd) < 0) goto error;
	        if (fputs("\n", fd) < 0) goto error;
            iCol = 0;
        }

        // Finish off.
        if (fputs("</pre></body>\n</html>\n", fd) < 0) goto error;

        fclose(fd);
        return;
    }

report:
    MessageBox(g_hwndMain, _T("Problems creating the HTML file!"), _T("C2HTML"), MB_OK|MB_ICONSTOP);
    return;

error:
    fclose(fd);
    remove(pszFilename);
    goto report;
}

/****************************************************************************
 *                                                                          *
 * Function: HTML_WriteText                                                 *
 *                                                                          *
 * Purpose : Output a HTML string - look for impopular characters.          *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-16  Created                                              *
 *                                                                          *
 ****************************************************************************/

#define TAB_STEP  4
static BOOL HTML_WriteText(FILE *fd, PCSTR pchText, int cchText, PINT piCol)
{
    PCSTR pch, pchEnd;

    // Walk through the text block.
    for (pch = pchText, pchEnd = pchText + cchText; pch < pchEnd; pch++)
    {
        switch (*pch)
        {
            case '\t':
            {
                // Output spaces to next tab-stop.
                int cSpaces = TAB_STEP - (*piCol) % TAB_STEP;
                do
                {
                    if (fputs(" ", fd) < 0)
                        return FALSE;
                    (*piCol)++;
                } while (--cSpaces > 0);
                break;
            }

            case ' ':
                // Blank at the beginning, or several blanks in a row?
                if (pch == pchText || (pch+1) < pchEnd && *(pch+1) == ' ')
                {
                    for ( ; pch < pchEnd && *pch == ' '; pch++)
                    {
                        if (fputs(" ", fd) < 0)
                            return FALSE;
                        (*piCol)++;
                    }
                    if (pch == pchEnd) continue;
                }
                /* fall through */
            default:
                if (fputc(*pch, fd) < 0)
                    return FALSE;
                (*piCol)++;
                break;

            case '<':
                if (fputs("&lt;", fd) < 0)
                    return FALSE;
                (*piCol)++;
                break;

            case '>':
                if (fputs("&gt;", fd) < 0)
                    return FALSE;
                (*piCol)++;
                break;

            case '"':
                if (fputs("&quot;", fd) < 0)
                    return FALSE;
                (*piCol)++;
                break;

            case '&':
                if (fputs("&amp;", fd) < 0)
                    return FALSE;
                (*piCol)++;
                break;

            case '^':
                if (fputs("&circ;", fd) < 0)
                    return FALSE;
                (*piCol)++;
                break;

            case '~':
                if (fputs("&tilde;", fd) < 0)
                    return FALSE;
                (*piCol)++;
                break;
        }
    }

    return TRUE;
}
#undef TAB_STEP

/****************************************************************************
 *                                                                          *
 * Function: ExportPDF                                                      *
 *                                                                          *
 * Purpose : Export C source code as PDF file.                              *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void ExportPDF(PCSTR pszFilename, PCTSTR pszText)
{
    PPDFFILE pPDF = NULL;

    FILE *fd = fopen(pszFilename, "wb");
    if (fd != NULL)
    {
        PPDFOBJ pObj;
        UINT uCookie;
        PCTSTR pchLine, pchNext;
        int nLineNo = 0;

        // Initialize.
        if ((pPDF = PDF_Init()) == NULL)
            goto error;

        // Start new page.
        if ((pObj = PDF_StartPage(pPDF)) == NULL)
            goto error;

        // Walk through the source file.
        uCookie = 0;
        for (pchLine = pchNext = pszText; *pchLine != '\0'; pchLine = pchNext)
        {
            int cPoints = 0, iPoint, iColor;
            PCTSTR pchCode, pchText;
            static PARSEPOINT aPoints[4096];

            // Find the end of the current line.
            while (*pchNext != '\0' && *pchNext++ != '\n')
                ;

            // Parse the current line, to get syntax color points.
            uCookie = C_Parser(uCookie, pchLine, (pchNext - pchLine), aPoints, &cPoints);

            // Walk through the line again, looking for color changes.
            iPoint = 0; iColor = COLOR_TEXT;
            for (pchCode = pchText = pchLine; pchText < pchNext; pchText++)
            {
                int iChar = pchText - pchLine;

                // Hit EOL?
                if (*pchText == '\n')
                    break;

                // Hit a color change?
                if (iPoint < cPoints && aPoints[iPoint].iChar == iChar)
                {
                    // Anything to print before the color change?
                    if (pchCode < pchText)
                    {
                        // Append color and text to page.
                        if (PDF_EmitColor(pObj, iColor) < 0 ||
                            PDF_EmitText(pObj, pchCode, (pchText - pchCode)) < 0)
                            goto error;

                        pchCode = pchText;
                    }

                    // In case we have multiple changes in the same spot...
                    while (iPoint < cPoints && aPoints[iPoint].iChar == iChar)
                        iColor = aPoints[iPoint++].iColor;
                }
            }

            // Handle any dangling text.
            if (pchCode < pchText)
            {
                // Append color and text to page.
                if (PDF_EmitColor(pObj, iColor) < 0 ||
                    PDF_EmitText(pObj, pchCode, (pchText - pchCode)) < 0)
                    goto error;
            }

            // Why not a line break?
            if (PDF_NewLine(pObj) < 0)
                goto error;

            if (++nLineNo > PDF_MAXLINES && *pchNext != '\0')
            {
                // Finalize current page.
                if (PDF_EndPage(pObj) < 0)
                    goto error;

                // Start new page.
                if ((pObj = PDF_StartPage(pPDF)) == NULL)
                    goto error;

                nLineNo = 0;
            }
        }

        // Finalize last page.
        if (PDF_EndPage(pObj) < 0)
            goto error;

        // Write the PDF file.
        if (PDF_WriteFile(pPDF, fd) < 0)
            goto error;

        // Clean up.
        PDF_Done(pPDF);

        fclose(fd);
        return;
    }

report:
    MessageBox(g_hwndMain, _T("Problems creating the PDF file!"), _T("C2HTML"), MB_OK|MB_ICONSTOP);
    return;

error:
    fclose(fd);
    remove(pszFilename);
    if (pPDF) PDF_Done(pPDF);
    goto report;
}

/****************************************************************************
 *                                                                          *
 * Function: PDF_Init                                                       *
 *                                                                          *
 * Purpose : Initialize PDF File output.                                    *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *                                                                          *
 ****************************************************************************/

static PPDFFILE PDF_Init(void)
{
    PPDFFILE pPDF = malloc(sizeof(*pPDF));
    if (pPDF != NULL)
    {
        PPDFOBJ pObj;

        pPDF->cObjects = 0;
        pPDF->pObjHead = NULL;
        pPDF->pObjProc = NULL;
        pPDF->pObjFont = NULL;

        // Allocate Document Catalog.
        if (!PDF_NewObject(pPDF, PDF_TYPE_CATALOG))
            return NULL;

        // Allocate Outlines Dictionary.
        if (!PDF_NewObject(pPDF, PDF_TYPE_OUTLINES))
            return NULL;

        // Allocate Pages Dictionary.
        if (!PDF_NewObject(pPDF, PDF_TYPE_PAGES))
            return NULL;

        // Allocate Procedure Set node.
        if (!(pPDF->pObjProc = PDF_NewObject(pPDF, PDF_TYPE_PROCSET)))
            return NULL;

        if (PDF_AppendData(pPDF->pObjProc, "[/PDF /Text]\n") < 0)
            return NULL;

        // Allocate Font node.
        if (!(pPDF->pObjFont = PDF_NewObject(pPDF, PDF_TYPE_FONT)))
            return NULL;

        if (PDF_AppendData(pPDF->pObjFont, "/Type /Font /Subtype /Type1 /Name /F1 "
            "/BaseFont /Courier /Encoding /WinAnsiEncoding\n") < 0)
            return NULL;

        // Allocate Info node.
        if (!(pObj = PDF_NewObject(pPDF, PDF_TYPE_INFO)))
            return NULL;

        if (PDF_AppendData(pObj, "/Producer (Export Add-In - www.smorgasbordet.com)\n") < 0)
            return NULL;
    }
    return pPDF;
}

/****************************************************************************
 *                                                                          *
 * Function: PDF_Done                                                       *
 *                                                                          *
 * Purpose : Terminate PDF File output.                                     *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void PDF_Done(PPDFFILE pPDF)
{
    // Walk through all object nodes.
    for (PPDFOBJ pObj = pPDF->pObjHead, pObjNext; pObj != NULL; pObj = pObjNext)
    {
        // Get next node before free().
        pObjNext = pObj->pNext;
        free(pObj->pszData);
        free(pObj);
    }

    free(pPDF);
}

/****************************************************************************
 *                                                                          *
 * Function: PDF_NewObject                                                  *
 *                                                                          *
 * Purpose : Add a new object node to the PDF file.                         *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *                                                                          *
 ****************************************************************************/

static PPDFOBJ PDF_NewObject(PPDFFILE pPDF, PDF_TYPE nType)
{
    PPDFOBJ pObj = malloc(sizeof(*pObj));
    if (pObj != NULL)
    {
        pObj->pNext = NULL;
        pObj->nType = nType;
        pObj->nNumber = ++pPDF->cObjects;
        pObj->nFileOffset = 0;
        pObj->pszData = NULL;

        if (!pPDF->pObjHead)
        {
            // This is the first node. Start the list.
            pPDF->pObjHead = pObj;
        }
        else
        {
            PPDFOBJ pObjT;

            // Find the end of the list and tack on the new node.
            for (pObjT = pPDF->pObjHead; pObjT->pNext; pObjT = pObjT->pNext)
                    ;

            pObjT->pNext = pObj;
        }
    }
    return pObj;
}

/****************************************************************************
 *                                                                          *
 * Function: PDF_FindObject                                                 *
 *                                                                          *
 * Purpose : Find object node by type.                                      *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *                                                                          *
 ****************************************************************************/

static PPDFOBJ PDF_FindObject(PPDFOBJ pObjHead, PDF_TYPE nType)
{
    // Walk through all object nodes.
    for (PPDFOBJ pObj = pObjHead; pObj != NULL; pObj = pObj->pNext)
        if (pObj->nType == nType) return pObj;

    return NULL;
}

/****************************************************************************
 *                                                                          *
 * Function: PDF_AppendData                                                 *
 *                                                                          *
 * Purpose : Add data to an object node.                                    *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *                                                                          *
 ****************************************************************************/

static int PDF_AppendData(PPDFOBJ pObj, PCSTR pszText)
{
    PSTR pszData;

    if (pObj->pszData != NULL)
        // KISS: Use realloc() for each call/block.
        pszData = realloc(pObj->pszData, strlen(pObj->pszData) + strlen(pszText) + 1);
    else
        // Use calloc() to clear.
        pszData = calloc(1, strlen(pszText) + 1);

    if (pszData == NULL)
        return -1;

    // Append data.
    pObj->pszData = strcat(pszData, pszText);
    return 0;
}

/****************************************************************************
 *                                                                          *
 * Function: PDF_StartPage                                                  *
 *                                                                          *
 * Purpose : Start a page; Allocate Page and Content nodes.                 *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *                                                                          *
 ****************************************************************************/

static PPDFOBJ PDF_StartPage(PPDFFILE pPDF)
{
    PPDFOBJ pObj;

    // Allocate Page and Page Content nodes.
    if ((pObj = PDF_NewObject(pPDF, PDF_TYPE_PAGE)) != NULL &&
        (pObj = PDF_NewObject(pPDF, PDF_TYPE_CONTENT)) != NULL)
    {
        // Initialize Page Content node
        if (PDF_AppendData(pObj, "BT\n1 0 0 1 50 760 Tm\n/F1 10 Tf\n0 -12.0 TD\n") < 0)
            return NULL;

        pObj->iCol = 0;
        pObj->iLastColor = -1;
    }

    return pObj;
}

/****************************************************************************
 *                                                                          *
 * Function: PDF_EndPage                                                    *
 *                                                                          *
 * Purpose : End a page; Finalize things.                                   *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *                                                                          *
 ****************************************************************************/

static int PDF_EndPage(PPDFOBJ pObj)
{
    // Finalize Page Content node.
    return PDF_AppendData(pObj, "ET\n");
}

/****************************************************************************
 *                                                                          *
 * Function: PDF_NewLine                                                    *
 *                                                                          *
 * Purpose : Emit a line-break.                                             *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *                                                                          *
 ****************************************************************************/

static int PDF_NewLine(PPDFOBJ pObj)
{
    pObj->iCol = 0;
    return PDF_AppendData(pObj, "T*\n");
}

/****************************************************************************
 *                                                                          *
 * Function: PDF_WriteFile                                                  *
 *                                                                          *
 * Purpose : Write the PDF file (when we know enough to do so).             *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *           05-03-21  Added FlateDecode filter (thanks Timppa!).           *
 *                                                                          *
 ****************************************************************************/

static int PDF_WriteFile(PPDFFILE pPDF, FILE *fd)
{
    long nOffset;

    // Write Header.
    if (fputs("%PDF-1.4\n", fd) < 0)
        return -1;

#ifdef USE_PDF_COMPRESSION
    // Indicate binary data in the file.
    if (fputs("%\xE2\xE3\xCF\xD3\n\n", fd) < 0)
        return -1;
#endif /* USE_PDF_COMPRESSION */

    // Walk through all object nodes, i.e. write the body.
    for (PPDFOBJ pObj = pPDF->pObjHead; pObj != NULL; pObj = pObj->pNext)
    {
        // Remember offset for cross-reference table.
        pObj->nFileOffset = ftell(fd);

        switch (pObj->nType)
        {
            // Write Document Catalog.
            case PDF_TYPE_CATALOG:
            {
                if (fprintf(fd, "%d 0 obj\n<<\n", pObj->nNumber) < 0)
                    return -1;

                if (fprintf(fd, "/Type /Catalog /Outlines %d 0 R /Pages %d 0 R\n",
                    PDF_FindObject(pPDF->pObjHead, PDF_TYPE_OUTLINES)->nNumber,
                    PDF_FindObject(pPDF->pObjHead, PDF_TYPE_PAGES)->nNumber) < 0)
                    return -1;

                if (fputs(">>\nendobj\n\n", fd) < 0)
                    return -1;
                break;
            }
            // Write Outline Dictionary.
            case PDF_TYPE_OUTLINES:
            {
                if (fprintf(fd, "%d 0 obj\n<<\n", pObj->nNumber) < 0)
                    return -1;

                if (fputs("/Type /Outlines /Count 0\n", fd) < 0)
                    return -1;

                if (fputs(">>\nendobj\n\n", fd) < 0)
                    return -1;
                break;
            }
            // Write Pages Dictionary.
            case PDF_TYPE_PAGES:
            {
                int cPages = 0;

                if (fprintf(fd, "%d 0 obj\n<<\n", pObj->nNumber) < 0)
                    return -1;

                if (fputs("/Type /Pages /Kids [\n", fd) < 0)
                    return -1;

                // Walk through all object nodes; Look for Page nodes.
                for (PPDFOBJ pObjT = pPDF->pObjHead; pObjT != NULL; pObjT = pObjT->pNext)
                {
                    if (pObjT->nType == PDF_TYPE_PAGE)
                    {
                        if (fprintf(fd, "%d 0 R\n", pObjT->nNumber) < 0)
                            return -1;

                        // Count pages.
                        cPages++;
                    }
                }

                if (fprintf(fd, "] /Count %d\n", cPages) < 0)
                    return -1;

                if (fputs(">>\nendobj\n\n", fd) < 0)
                    return -1;
                break;
            }
            // Write single Page Node.
            case PDF_TYPE_PAGE:
            {
                if (fprintf(fd, "%d 0 obj\n<<\n", pObj->nNumber) < 0)
                    return -1;

                if (fprintf(fd, "/Type /Page /Parent %d 0 R /MediaBox [0 0 612 792] /Contents %d 0 R\n",
                    PDF_FindObject(pPDF->pObjHead, PDF_TYPE_PAGES)->nNumber,
                    pObj->pNext->nNumber /* Content is always next node */) < 0)
                    return -1;

                if (fprintf(fd, "/Resources << /ProcSet %d 0 R /Font << /F1 %d 0 R >> >>\n",
                    pPDF->pObjProc->nNumber,
                    pPDF->pObjFont->nNumber) < 0)
                    return -1;

                if (fputs(">>\nendobj\n\n", fd) < 0)
                    return -1;
                break;
            }
            // Write Procedure Set node.
            case PDF_TYPE_PROCSET:
            {
                if (fprintf(fd, "%d 0 obj\n", pObj->nNumber) < 0)
                    return -1;

                if (fputs(pObj->pszData, fd) < 0)
                    return -1;

                if (fputs("endobj\n\n", fd) < 0)
                    return -1;
                break;
            }
            // Write Font node.
            case PDF_TYPE_FONT:
            {
                if (fprintf(fd, "%d 0 obj\n<<\n", pObj->nNumber) < 0)
                    return -1;

                if (fputs(pObj->pszData, fd) < 0)
                    return -1;

                if (fputs(">>\nendobj\n\n", fd) < 0)
                    return -1;
                break;
            }
            // Write Page Content node.
            case PDF_TYPE_CONTENT:
            {
                if (pObj->pszData == NULL)  /* avoid disaster */
                    pObj->pszData = calloc(1,1);

#ifdef USE_PDF_COMPRESSION
                PBYTE pCmpr = NULL;

                __try
                {
                    unsigned long cbData = strlen(pObj->pszData);
                    unsigned long cbCmpr = compressBound(cbData);

                    if ((pCmpr = malloc(cbCmpr)) == NULL)
                        return -1;

                    if (compress2(pCmpr, &cbCmpr, (PBYTE)pObj->pszData, cbData, Z_DEFAULT_COMPRESSION) != Z_OK)
                        return -1;

                    if (fprintf(fd, "%d 0 obj\n<<\n", pObj->nNumber) < 0)
                        return -1;

                    if (fprintf(fd, "/Length %lu /Filter /FlateDecode\n>>\nstream\n", cbCmpr) < 0)
                        return -1;

                    if (fwrite(pCmpr, cbCmpr, 1, fd) < 1)
                        return -1;
                }
                __finally
                {
                    // Works for NULL too.
                    free(pCmpr);
                }
#else /* USE_PDF_COMPRESSION */
                if (fprintf(fd, "%d 0 obj\n<<\n", pObj->nNumber) < 0)
                    return -1;

                if (fprintf(fd, "/Length %d\n>>\nstream\n", strlen(pObj->pszData)) < 0)
                    return -1;

                if (fputs(pObj->pszData, fd) < 0)
                    return -1;

#endif /* USE_PDF_COMPRESSION */
                if (fputs("\nendstream\nendobj\n\n", fd) < 0)
                    return -1;
                break;
            }
            // Write Info node.
            case PDF_TYPE_INFO:
            {
                if (fprintf(fd, "%d 0 obj\n<<\n", pObj->nNumber) < 0)
                    return -1;

                if (fputs(pObj->pszData, fd) < 0)
                    return -1;

                if (fputs(">>\nendobj\n\n", fd) < 0)
                    return -1;
                break;
            }
        }
    }

    // Remember offset for file trailer.
    nOffset = ftell(fd);

    // 
    // Write cross-reference table.
    //
    if (fprintf(fd, "xref\n0 %d\n", pPDF->cObjects + 1) < 0)
        return -1;

    // Always 20 bytes long; either end with " \n" or "\r\n".
    if (fputs("0000000000 65535 f \n", fd) < 0)
        return -1;

    for (PPDFOBJ pObj = pPDF->pObjHead; pObj != NULL; pObj = pObj->pNext)
    {
        // Always 20 bytes long; either end with " \n" or "\r\n".
        if (fprintf(fd, "%010ld 00000 n \n", pObj->nFileOffset) < 0)
            return -1;
    }

    // 
    // Write File Trailer.
    //
    if (fputs("\ntrailer\n<<\n", fd) < 0)
        return -1;

    if (fprintf(fd, "/Size %d /Root %d 0 R /Info %d 0 R\n",
        pPDF->cObjects + 1 /* number of entries in cross-reference table */,
        PDF_FindObject(pPDF->pObjHead, PDF_TYPE_CATALOG)->nNumber,
        PDF_FindObject(pPDF->pObjHead, PDF_TYPE_INFO)->nNumber) < 0)
        return -1;

    if (fprintf(fd, ">>\nstartxref\n%ld\n", nOffset) < 0)
        return -1;

    // Write EOF marker.
    if (fputs("%%EOF\n", fd) < 0)
        return -1;

    return 0;
}

/****************************************************************************
 *                                                                          *
 * Function: PDF_EmitText                                                   *
 *                                                                          *
 * Purpose : Emit text to PDF text stream - look for impopular characters.  *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *                                                                          *
 ****************************************************************************/

#define TAB_STEP  4
static int PDF_EmitText(PPDFOBJ pObj, PCSTR pchText, int cchText)
{
    static CHAR buf[256];
    PCSTR pchEnd = pchText + cchText;
    PCSTR pch;
    PSTR pchOut;

    pchOut = buf;
    *pchOut++ = '(';

    // Walk through the text block.
    for (pch = pchText, pchEnd = pchText + cchText; pch < pchEnd && pchOut < &buf[NELEMS(buf)-10]; pch++)
    {
        // Just chop off, if line is too long. Not so nice, but have to work for now.
        if (pObj->iCol == PDF_MAXCHARS)
        {
            *pchOut++ = ')';
            *pchOut++ = ' ';
            *pchOut++ = 'T';
            *pchOut++ = 'j';
            *pchOut++ = '\n';
            *pchOut = '\0';

            if (PDF_AppendData(pObj, buf) < 0 || PDF_NewLine(pObj) < 0)
                return -1;

            pchOut = buf;
            *pchOut++ = '(';
        }

        switch (*pch)
        {
            case '\t':
            {
                // Output spaces to next tab-stop.
                int cSpaces = TAB_STEP - pObj->iCol % TAB_STEP;
                do
                {
                    *pchOut++ = ' ';
                    pObj->iCol++;
                } while (--cSpaces > 0);
                break;
            }

            case '(':
                *pchOut++ = '\\';
                *pchOut++ = '(';
                pObj->iCol++;
                break;

            case ')':
                *pchOut++ = '\\';
                *pchOut++ = ')';
                pObj->iCol++;
                break;

            case '\\':
                *pchOut++ = '\\';
                *pchOut++ = '\\';
                pObj->iCol++;
                break;

            default:
                *pchOut++ = *pch;
                pObj->iCol++;
                break;
        }
    }

    *pchOut++ = ')';
    *pchOut++ = ' ';
    *pchOut++ = 'T';
    *pchOut++ = 'j';
    *pchOut++ = '\n';
    *pchOut = '\0';

    return PDF_AppendData(pObj, buf);
}
#undef TAB_STEP

/****************************************************************************
 *                                                                          *
 * Function: PDF_EmitColor                                                  *
 *                                                                          *
 * Purpose : Emit color to PDF text stream (RGB format).                    *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-18  Created                                              *
 *                                                                          *
 ****************************************************************************/

static int PDF_EmitColor(PPDFOBJ pObj, int iColor)
{
    PCSTR pszColor = "";

    if (pObj->iLastColor == iColor)
        return 0;

    switch (iColor)
    {
        case COLOR_TEXT: pszColor = "0 0 0 rg "; break;
        case COLOR_KEYWORD: pszColor = "0 0 1.0 rg "; break;
        case COLOR_COMMENT: pszColor = "0 0.5 0 rg "; break;
        case COLOR_NUMBER: pszColor = "1.0 0 1.0 rg "; break;
        case COLOR_STRING: pszColor = "0.5 0 0.5 rg "; break;
        case COLOR_PREPROCESSOR: pszColor = "0.5 0.0 0.0 rg "; break;
        case COLOR_OPERATOR: pszColor = "0 0 1.0 rg "; break;
    }

    pObj->iLastColor = iColor;
    return PDF_AppendData(pObj, pszColor);
}

/****************************************************************************
 *                                                                          *
 * Function: C_Parser                                                       *
 *                                                                          *
 * Purpose : Syntax color parser.                                           *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-16  Created                                              *
 *                                                                          *
 ****************************************************************************/

static UINT C_Parser(UINT uCookie, PCTSTR pchText, int cchText, PARSEPOINT aPoints[], PINT pcPoints)
{
    BOOL fFirstChar;
    int iRedefine = 0;
    int iIdentBegin = -1;
    int i;

    fFirstChar = (uCookie & ~COOKIE_EXT_COMMENT) == 0;

    if (cchText == 0)
        return uCookie;

    for (i = 0;; i++)
    {
        if (iRedefine >= 0)
        {
            if (uCookie & (COOKIE_COMMENT|COOKIE_EXT_COMMENT))
            {
                DEFINE_BLOCK(iRedefine, COLOR_COMMENT);
            }
            else if (uCookie & (COOKIE_CHAR|COOKIE_STRING))
            {
                DEFINE_BLOCK(iRedefine, COLOR_STRING);
            }
            else if (uCookie & COOKIE_PREPROCESSOR)
            {
                DEFINE_BLOCK(iRedefine, COLOR_PREPROCESSOR);
            }
            else
            {
                DEFINE_BLOCK(iRedefine, COLOR_TEXT);
            }

            iRedefine = -1;
        }

        // check for end of line
        if (i == cchText)
            break;

        // inside single line comment: //
        if (uCookie & COOKIE_COMMENT)
        {
            DEFINE_BLOCK(i, COLOR_COMMENT);
            uCookie |= COOKIE_COMMENT;
            break;
        }

        // inside string constant: "...."
        if (uCookie & COOKIE_STRING)
        {
            // check for end of string constant...
            if (pchText[i] == '"' && (i == 0 || pchText[i-1] != '\\' || i > 2 && pchText[i-2] == '\\' && pchText[i-3] != '\\'))
            {
                uCookie &= ~COOKIE_STRING;
                iRedefine = i + 1;
            }
            continue;
        }

        // inside char constant: '.'
        if (uCookie & COOKIE_CHAR)
        {
            // check for end of char constant...
            if (pchText[i] == '\'' && (i == 0 || pchText[i-1] != '\\' || i > 1 && pchText[i-2] == '\\' && pchText[i-3] != '\\'))
            {
                uCookie &= ~COOKIE_CHAR;
                iRedefine = i + 1;
            }
            continue;
        }

        // inside extended comment: /*....*/
        if (uCookie & COOKIE_EXT_COMMENT)
        {
            // check for end of extended comment...
            if (i > 0 && pchText[i] == '/' && pchText[i-1] == '*')
            {
                uCookie &= ~COOKIE_EXT_COMMENT;
                iRedefine = i + 1;
            }
            continue;
        }

        // inside preprocessor directive: #...
        if (uCookie & COOKIE_PREPROCESSOR)
        {
            // check for string constant: "...."
            if (pchText[i] == '"')
            {
                uCookie |= COOKIE_STRING;
                continue;
            }
            // check for char constant: '.'
            else if (pchText[i] == '\'')
            {
                uCookie |= COOKIE_CHAR;
                continue;
            }
            // check for single line comment: //
            else if (i > 0 && pchText[i] == '/' && pchText[i-1] == '/')
            {
                DEFINE_BLOCK(i-1, COLOR_COMMENT);
                uCookie |= COOKIE_COMMENT;
                break;
            }
            // check for extended comment...
            else if (i > 0 && pchText[i] == '*' && pchText[i-1] == '/')
            {
                DEFINE_BLOCK(i-1, COLOR_COMMENT);
                uCookie |= COOKIE_EXT_COMMENT;
            }
            continue;
        }

        // check for single line comment: //
        if (i > 0 && pchText[i] == '/' && pchText[i-1] == '/')
        {
            DEFINE_BLOCK(i-1, COLOR_COMMENT);
            uCookie |= COOKIE_COMMENT;
            continue;
        }

        // check for string constant: "...."
        if (pchText[i] == '"')
        {
            DEFINE_BLOCK(i, COLOR_STRING);
            uCookie |= COOKIE_STRING;
            continue;
        }

        // check for char constant: '.'
        if (pchText[i] == '\'')
        {
            DEFINE_BLOCK(i, COLOR_STRING);
            uCookie |= COOKIE_CHAR;
            continue;
        }

        // check for extended comment: /*....*/
        if (i > 0 && pchText[i] == '*' && pchText[i-1] == '/')
        {
            DEFINE_BLOCK(i-1, COLOR_COMMENT);
            uCookie |= COOKIE_EXT_COMMENT;
            continue;
        }

        if (fFirstChar)
        {
            if (pchText[i] == '#')
            {
                DEFINE_BLOCK(i, COLOR_PREPROCESSOR);
                uCookie |= COOKIE_PREPROCESSOR;
                continue;
            }

            if (!isspace(pchText[i]))
                fFirstChar = FALSE;
        }

        // handle identifiers and keywords (and floating point numbers).
        if (isalnum(pchText[i]) || pchText[i] == '_' || ((pchText[i] == '.' || pchText[i] == '-' || pchText[i] == '+') &&
            iIdentBegin != -1 && IsNumber(pchText + iIdentBegin, i - iIdentBegin + 1)))
        {
            if (iIdentBegin == -1)
                iIdentBegin = i;
        }
        else
        {
            if (iIdentBegin >= 0)
            {
                if (IsKeyword(apszKeywords, NELEMS(apszKeywords), pchText + iIdentBegin, i - iIdentBegin))
                {
                    DEFINE_BLOCK(iIdentBegin, COLOR_KEYWORD);
                }
                else if (IsNumber(pchText + iIdentBegin, i - iIdentBegin))
                {
                    DEFINE_BLOCK(iIdentBegin, COLOR_NUMBER);
                }

                iRedefine = i;
                iIdentBegin = -1;
            }

            // check for operators.
            if (i > 1 && (
                (pchText[i-2] == '<' && pchText[i-1] == '<' && pchText[i] == '=') ||
                (pchText[i-2] == '>' && pchText[i-1] == '>' && pchText[i] == '=') ||
                (pchText[i-2] == '.' && pchText[i-1] == '.' && pchText[i] == '.')))
            {
                DEFINE_BLOCK(i-2, COLOR_OPERATOR);
                iRedefine = i + 1;
                continue;
            }
            else if (i > 0 && (
                (pchText[i-1] == '+' && pchText[i] == '+') ||
                (pchText[i-1] == '-' && pchText[i] == '-') ||
                (pchText[i-1] == '-' && pchText[i] == '>') ||
                (pchText[i-1] == '<' && pchText[i] == '<') ||
                (pchText[i-1] == '>' && pchText[i] == '>') ||
                (pchText[i-1] == '<' && pchText[i] == '=') ||
                (pchText[i-1] == '>' && pchText[i] == '=') ||
                (pchText[i-1] == '=' && pchText[i] == '=') ||
                (pchText[i-1] == '!' && pchText[i] == '=') ||
                (pchText[i-1] == '&' && pchText[i] == '&') ||
                (pchText[i-1] == '|' && pchText[i] == '|') ||
                (pchText[i-1] == '*' && pchText[i] == '=') ||
                (pchText[i-1] == '/' && pchText[i] == '=') ||
                (pchText[i-1] == '%' && pchText[i] == '=') ||
                (pchText[i-1] == '+' && pchText[i] == '=') ||
                (pchText[i-1] == '-' && pchText[i] == '=') ||
                (pchText[i-1] == '&' && pchText[i] == '=') ||
                (pchText[i-1] == '^' && pchText[i] == '=') ||
                (pchText[i-1] == '|' && pchText[i] == '=')))
            {
                DEFINE_BLOCK(i-1, COLOR_OPERATOR);
                iRedefine = i + 1;
                continue;
            }
            else if (
                pchText[i] == '(' ||
                pchText[i] == ')' ||
                pchText[i] == '[' ||
                pchText[i] == ']' ||
                pchText[i] == '{' ||
                pchText[i] == '}' ||
                pchText[i] == ',' ||
                (pchText[i] == '.' && ((i+1) == cchText || (pchText[i+1] != '.'))) ||
                (pchText[i] == '!' && ((i+1) == cchText || (pchText[i+1] != '='))) ||
                (pchText[i] == '~' && ((i+1) == cchText || (pchText[i+1] != '='))) ||
                (pchText[i] == '^' && ((i+1) == cchText || (pchText[i+1] != '='))) ||
                (pchText[i] == '*' && ((i+1) == cchText || (pchText[i+1] != '='))) ||
                (pchText[i] == '%' && ((i+1) == cchText || (pchText[i+1] != '='))) ||
                (pchText[i] == '=' && ((i+1) == cchText || (pchText[i+1] != '='))) ||
                (pchText[i] == '+' && ((i+1) == cchText || (pchText[i+1] != '=' && pchText[i+1] != '+'))) ||
                (pchText[i] == '-' && ((i+1) == cchText || (pchText[i+1] != '=' && pchText[i+1] != '-' && pchText[i+1] != '>'))) ||
                (pchText[i] == '&' && ((i+1) == cchText || (pchText[i+1] != '=' && pchText[i+1] != '&'))) ||
                (pchText[i] == '|' && ((i+1) == cchText || (pchText[i+1] != '=' && pchText[i+1] != '|'))) ||
                (pchText[i] == '/' && ((i+1) == cchText || (pchText[i+1] != '=' && pchText[i+1] != '/' && pchText[i+1] != '*'))) ||
                (pchText[i] == '<' && ((i+1) == cchText || (pchText[i+1] != '=' && pchText[i+1] != '<'))) ||
                (pchText[i] == '>' && ((i+1) == cchText || (pchText[i+1] != '=' && pchText[i+1] != '>'))))
            {
                DEFINE_BLOCK(i, COLOR_OPERATOR);
                iRedefine = i + 1;
                continue;
            }
        }
    }

    // handle dangling identifiers and keywords.
    if (iIdentBegin >= 0)
    {
        if (IsKeyword(apszKeywords, NELEMS(apszKeywords), pchText + iIdentBegin, i - iIdentBegin))
        {
            DEFINE_BLOCK(iIdentBegin, COLOR_KEYWORD);
        }
        else if (IsNumber(pchText + iIdentBegin, i - iIdentBegin))
        {
            DEFINE_BLOCK(iIdentBegin, COLOR_NUMBER);
        }
    }

    if (cchText < 2 || pchText[cchText-1] != '\n' || pchText[cchText-2] != '\\')
        uCookie &= COOKIE_EXT_COMMENT;

    return uCookie;
}

/****************************************************************************
 *                                                                          *
 * Function: IsKeyword                                                      *
 *                                                                          *
 * Purpose : Return TRUE if the current token is a keyword.                 *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-16  Created                                              *
 *                                                                          *
 ****************************************************************************/

static int IsKeyword(PCTSTR apszKeywords[], int cKeywords, PCTSTR pchText, int cchText)
{
    int lo = 0;
    int hi = cKeywords-1;

    while (hi >= lo)  /* binary search */
    {
        int this = (lo + hi) / 2;
        int cond = _tcsncmp(apszKeywords[this], pchText, cchText);

        if (cond > 0)
            hi = this-1;
        else if (cond < 0)
            lo = this+1;
        else if (apszKeywords[this][cchText] == '\0')
            return 1 + this;
        else if ((cond = IsKeyword(apszKeywords + lo, this - lo, pchText, cchText)) != 0)
            return cond + lo;
        else if ((cond = IsKeyword(apszKeywords + this + 1, hi - this, pchText, cchText)) != 0)
            return cond + this + 1;
        else
            break;
    }

    return 0;
}

/****************************************************************************
 *                                                                          *
 * Function: IsNumber                                                       *
 *                                                                          *
 * Purpose : Return TRUE if the current token is a number.                  *
 *                                                                          *
 * History : Date      Reason                                               *
 *           05-03-16  Created                                              *
 *                                                                          *
 ****************************************************************************/

static BOOL IsNumber(PCTSTR pchText, int cchText)
{
    BOOL fFloat = FALSE;
    int i;

    if (cchText > 2 && pchText[0] == '0' && pchText[1] == 'x')  /* hexadecimal */
    {
        i = 2;

        while (i < cchText && isxdigit(pchText[i]))
            i++;

        if (i < cchText && pchText[i] == '.')
            i++, fFloat = TRUE;

        while (i < cchText && isxdigit(pchText[i]))
            i++;

        if (i < cchText && (pchText[i] == 'p' || pchText[i] == 'P'))
        {
            i++, fFloat = TRUE;

            if (i < cchText && (pchText[i] == '-' || pchText[i] == '+'))
                i++;

            while (i < cchText && _istdigit(pchText[i]))
                i++;
        }
    }
    else if (_istdigit(pchText[0]))  /* decimal */
    {
        i = 1;

        while (i < cchText && _istdigit(pchText[i]))
            i++;

        if (i < cchText && pchText[i] == '.')
            i++, fFloat = TRUE;

        while (i < cchText && _istdigit(pchText[i]))
            i++;

        if (i < cchText && (pchText[i] == 'e' || pchText[i] == 'E'))
        {
            i++, fFloat = TRUE;

            if (i < cchText && (pchText[i] == '-' || pchText[i] == '+'))
                i++;

            while (i < cchText && _istdigit(pchText[i]))
                i++;
        }
    }
    else return FALSE;

    if (fFloat)
    {
        if (i < cchText && _tcschr(_T("fFlL"), pchText[i]) != NULL)
            i++;
    }
    else
    {
        while (i < cchText && _tcschr(_T("uUlL"), pchText[i]) != NULL)
            i++;
    }

    return (i < cchText) ? FALSE : TRUE;
}
